package Net::OICQ;

# $Id: OICQ.pm,v 1.30 2003/10/17 20:01:52 tans Exp $

# Copyright (c) 2002 Shufeng Tan.  All rights reserved.
# 
# This package is free software and is provided "as is" without express
# or implied warranty.  It may be used, redistributed and/or modified
# under the terms of the Perl Artistic License (see
# http://www.perl.com/perl/misc/Artistic.html)

use 5.006;
use strict;
use warnings;
use Carp;
use FileHandle;
use Digest::MD5;
use IO::Socket;
eval "no encoding; use bytes;" if $] >= 5.008;

use Crypt::OICQ qw(encrypt decrypt);

our $VERSION = '0.8';

#################### Begin OICQ protocol data ######################

# Byte 0x00 is always 0x02, i suspect it is protocol version

our $PROTO_VER = "\x02";

# Bytes 0x01-0x02 indicates packet source

# 0x01 0x00 for packets from server
# 0x06 0x2e for packets from GB client version 2000c build 630
# 0x07 0x2e for packets from En client version 2000c build 305
# 0x08 0x01 for packets from En client version 2000c build 630
# 0x09 0x09 for packets from GB client version 2000c build 1230b

our $SRC_SERVER = "\x01\x00";
our $SRC_CLIENT = "\x08\x01"; # English version 2000c build 630

# Byte 0x03 is always 0x00, maybe reserved for future commands?
# Byte 0x04 is command

our %CmdCode = (
    logout             => "\0\x01",
    keep_alive         => "\0\x02",
    reg_new_id_2       => "\0\x03",
    update_info        => "\0\x04",
    search_users       => "\0\x05",
    get_user_info      => "\0\x06",
    add_contact_1      => "\0\x09",
    del_contact        => "\0\x0a",
    add_contact_2      => "\0\x0b",
    set_mode           => "\0\x0d",
    reg_new_id_1       => "\0\x11",
    ack_service_msg    => "\0\x12",
    send_msg           => "\0\x16",
    recv_msg           => "\0\x17",
    forbid_contact     => "\0\x1c",
    login              => "\0\x22",
    get_friends_list   => "\0\x26",
    get_online_friends => "\0\x27",
    recv_service_msg   => "\0\x80",
    recv_friend_status => "\0\x81",
);

# Bytes 0x05-0x06 form a packet sequence number, a 16-bit integer

# Message precursor

my $MsgPrecursor = "\x00\x00\x00\x00\x00\x01\x00\x00\x00\x00\x01";

# Login modes
our %ConnectMode = (
    Normal    => "\x0a",
    Invisible => "\x28"
);

# System message code for 0x80 cmd
our %ServiceMsgCode = (
    '01' => 'User',
    '02' => 'ContactRequest',
    '06' => 'Broadcast'
);

# Separators
our $FS = "\x1e";   # Field separator
our $RS = "\x1f";   # Record separator

our @InfoHeader = qw(
    UserID Nickname Country Province PostCode Street Phone Age Sex Realname
    Email PagerCode PagerProvider PagerStationNum PagerNum PagerType
    Occupation Homepage Authorization unkn19 unkn20 Avatar
    MobilePhone MobileType Aboutme City unkn26 unkn27 unkn28 PublishMobile
    PublishContact School Horoscope Shengxiao BloodType unkn35 unkn36
);

# Some constants for constructing client packets
my $PacketHead = $PROTO_VER . $SRC_CLIENT;
my $PacketTail = "\x03";

my $C_LOGIN_17_33 = pack('H*', join("", qw(
00 00 00 09 00 00 00 01 00 00 00 11 f3 c8 9b c5
1f 24 ce d2 9e fe de ba cc 52 b4 b3 02
)));

my $C_LOGIN_35_END = pack('H*', join("", qw(
4d ea b4 a9 e6 85 b0 46 bb 0b 66 a0 1a f4 b6 b2
)));

#################### End OICQ protocol data ########################

my @ServerIPs = qw(
    61.144.238.146     202.104.129.251    202.104.129.254     61.141.194.203
    202.104.129.252    202.104.129.253     61.144.238.145     218.17.209.22
);

# Constructor

sub new {
    my ($class) = @_;
    my $dir = "$ENV{HOME}/.oicq";
    if (-e $dir) {
        -d $dir or croak "$dir exists but is not a directory";
    } else {
        mkdir($dir) or croak "Failed to mkdir $dir: $!";
    }
    my $self = {
        Dir        => $dir,
        LastSvrAck => 0,
        Debug      => 0  # 1 - trace packets, 2 - desect packets
    };
    my $logfile = "$dir/oicq.log";
    my $log = new FileHandle ">>$logfile";
    defined($log) or croak "Failed to open >>$logfile";
    $log->autoflush;
    $self->{LogFile} = $logfile;
    $self->{Log} = $log;
    return bless($self, $class);
}

# Methods that do not require connection to a server

sub set_user {
    my ($self, $id, $pw) = @_;

    $self->{Id} = $id;
    $self->{Passwd} = $pw;
    $self->{_Id} = pack('N', $id);
    $self->{PWKey} = Digest::MD5::md5($pw);
    $self->{EventQueue} = [];
    $self->{SearchCount} = 0;
    $self->{LogChat}   = 1;
    $self->{Info}      = {};      # use id as hash key
    $self->{Away}      = 0;
    $self->{LastAutoReply} = {};  # use id as hash key
    $self->{AutoAwayTime} = "";

    my $userdir = "$self->{Dir}/$id";
    -e $userdir or mkdir($userdir);
    if (-d $userdir) {
        foreach ($self->get_saved_ids) { $self->get_nickname($_) };
        my $logfile = "$userdir/user.log";
        my $log = new FileHandle(">>$logfile");
        if (defined $log) {
            $self->log_t("Switch log to $logfile") if $self->{Debug};
            $self->{Log} = undef;
            $self->{LogFile} = $logfile;
            $self->{Log} = $log;
            $log->autoflush;
        } else {
            $self->log_t("Failed to open >>$logfile");
        }
    } else {
        $self->log_t("Failed to mkdir $userdir");
    }
}

# Methods for building OICQ packets

sub build_packet {
    my ($self, $cmd, $data) = @_;
    croak "build_packet error: bad command: $cmd" unless exists $CmdCode{$cmd};
    $self->{Seq}++;
    my $crypt = encrypt(undef, $data, $self->{Key});
    my $packet = $PacketHead . $CmdCode{$cmd} . pack('n', $self->{Seq}) .
                 $self->{_Id} . $crypt . $PacketTail;
    return $packet;
}

sub rand_str {
    my $len = pop;
    join('', map(pack("C", rand(0xff)), 1..$len));
}

sub build_login_packet {
    my ($self) = @_;
    my $seq = "\0" . rand_str(1);
    $self->{Seq} = unpack('n', $seq);

    my $randkey = rand_str(16);
    $self->{RandKey} = $randkey;

    # 0x00-0x10
    my $data = encrypt(undef, "", $self->{PWKey}) . "\0";
    # 0x11-0x22    We use fake ip 10.0.0.1
    $data .= pack('Nn', 0x0a000001, 4000);
    # 0x23-0x43
    $data .= $C_LOGIN_17_33 . $ConnectMode{$self->{ConnectMode}} .
             $C_LOGIN_35_END;

    $PacketHead . $CmdCode{'login'} . $seq . $self->{_Id} . $randkey .
    encrypt(undef, $data, $randkey) . $PacketTail;
}

sub build_logout_packet {
    my ($self) = @_;
    $PacketHead . $CmdCode{'logout'} . ("\xff" x 2) . $self->{_Id} .
    encrypt(undef, $self->{PWKey}, $self->{Key}) . $PacketTail;
}

# person to person packet not done

sub build_p2p_packet {
    my ($self, $dstid, $cmd, $data) = @_;
    $data;
}

# Tencent no longer allows getting new number this way
# So these two are pretty much useless

sub build_reg_1_packet {
    my ($self) = @_;
    my $randkey = rand_str(16);
    my $data = "";   # fixme
    $PacketHead . $CmdCode{'reg_new_id_1'} . pack('n', 1) . "\0"x4 .
    $randkey . encrypt(undef, $data, $randkey) . $PacketTail;
}

sub build_reg_2_packet {
    my ($self) = @_;
    my $randkey = rand_str(16);
    my $data = "";   # fixme
    $PacketHead . $CmdCode{'reg_new_id_2'} . pack('n', 2) . "\0"x4 .
    $randkey . encrypt(undef, $data, $randkey) . $PacketTail;
}

# Methods for logging and output

sub log {
    my $self = shift;
    my $log = $self->{Log};
    print $log @_;
}

sub logf {
    my $self = shift;
    my $log = $self->{Log};
    printf $log @_;
}

sub log_t {
    my ($self, @msg) = @_;
    my $log = $self->{Log};
    print $log substr(localtime, 4, 16), @msg, "\n";
}

sub hexdump {
    my $str = pop;
    return unless defined $str;
    my $res = "";
    my $len = length($str);
    for (my $i = 0; $i < $len; $i += 16) {
        my $s = substr($str, $i, 16);
        my $hex = unpack('H*', $s);
        $s =~ s/[\x00-\x1f]/./g;   # 0x00-0x1f will screw up terminal
        $hex =~ s/(\w\w)/$1 /g;
        $res .= sprintf("%-48s    %s\n", $hex, $s);
    }
    return $res;
}

sub dump_substr {
    my ($self, $data, $tmpl, $prefix, $begin, $len) = @_;
    my ($str, $end);
    if (defined($len)) {
        $str = substr($data, $begin, $len);
        $end = ($begin+$len < length($data)) ? $begin+$len-1 : length($data)-1;
    } else {
        $str = substr($data, $begin);
        $end = length($data)-1;
    }
    $self->logf("0x%02x-0x%02x %s: ", $begin, $end, $prefix);
    if ($tmpl =~ /\w/) {
        if ($tmpl eq 'H*') {
            $self->log("\n", $self->hexdump($str));
        } else {
            $self->log(unpack($tmpl, $str), "\n");
        }
    } else {
        $self->log("$str\n");
    }
}

sub desect {
    my $self = shift;
    return unless $self->{Debug} > 1;
    my $data = shift;
    foreach my $arg (@_) {
        $self->dump_substr($data, @{$arg});
    }
    return;
}

sub remove_saved_id {
    my ($self, $id) = @_;
    my $file = "$self->{Dir}/$self->{Id}/$id.dat";
    if (-e $file) {
        unlink($file);
        return 0 if -e $file;
        return 1;
    } else {
        return 0;
    }
}

sub get_saved_ids {
    my ($self) = @_;
    my $dir = "$self->{Dir}/$self->{Id}";
    my @ids = ();
    if (opendir(DIR, $dir)) {
        while(my $f = readdir(DIR)) {
            next unless $f =~ /^(\d+)\.dat$/;
            push @ids, $1;
        }
        closedir(DIR);
    }
    return @ids;
}

sub get_face {
    my $num = pop;
    sprintf('%d-%d', 1 + $num/3, 1 + $num % 3);
}

sub toggle_autoreply {
    my ($self) = @_;
    if ($self->{Away}) {
        $self->{Away} = 0;
        return "off";
    } else {
        $self->{Away} = 1;
        return "on";
    }
}

# Nickname can be updated by get_friends_list or get_user_info

sub get_nickname {
    my ($self, $id) = @_;
    if (defined $self->{Info}->{$id}) {
        if (defined $self->{Info}->{$id}->{Nickname}) {
            return $self->{Info}->{$id}->{Nickname};
        }
    } else {
        $self->{Info}->{$id} = {};
    }
    my $infofile = "$self->{Dir}/$self->{Id}/$id.dat";
    my $nick = "";
    if (open(INFO, $infofile)) {
        while(my $line = <INFO>) {
            if ($line =~ /^Nickname +=> *'(.*)'/) {
                $nick = $1;
                last;
            }
        }
        close(INFO);        
    }
    $self->{Info}->{$id}->{Nickname} = $nick;
    return $nick;
}

sub get_servers {
    my @servers;
    if (exists $ENV{OICQ_SVR} and $ENV{OICQ_SVR} =~ /\d+/) {
        my $svr = $ENV{OICQ_SVR};
        $svr =~ s/^\D+//;
        $svr =~ s/\D+$//;
        @servers = split(/[^\w\.]+/, $svr);
        return @servers if @servers;
    }

    eval "no encoding;" if $] >= 5.008;
    foreach my $n (2..7, "") {
        my $ip = gethostbyname("sz$n.tencent.com");
        next unless $ip;
        push @servers, join('.', map(ord, split('', $ip)));
    }
    if (@servers) {
        # this unregistered ip seems to be for mobile numbers only
        push @servers, '218.17.209.22';
        return @servers;
    }
    return @ServerIPs;
}

sub connect {
    my ($self, $ip, $port) = @_;
    croak "Server IP not provided"  unless defined($ip) && $ip;
    $port = 8000 unless defined($port) && $port;

    my $socket = IO::Socket::INET->new(
            Proto => 'udp', PeerAddr => $ip, PeerPort => $port
    ) or croak("socket error: $@");

    $self->{SvrIP}   = $ip;
    $self->{SvrPort} = $port;
    $self->{Socket}  = $socket;
    return $socket;
}

sub login {
    my ($self, $id, $pw, $mode) = @_;
    $self->set_user($id, $pw);
    $self->{Key} = ""; 

    if (defined $mode && exists $ConnectMode{$mode}) {
        $self->log_t("login as $id in $mode mode");
        $self->{ConnectMode} = $mode;
    } else {
        $self->log_t("login as $id, default to Normal mode");
        $self->{ConnectMode} = 'Normal';
    }
    my $packet = $self->build_login_packet;
    my @servers = $self->get_servers();
    my $port = 8000;
    my $timeout = 30;
    my $timeout_msg = "tImEoUt\n";
 LOGIN: foreach my $svr (@servers) {
        print "Connecting to server $svr:$port...";
        my $socket = $self->connect($svr, $port);
        $socket->send($packet) && print "...";
        my $resp = undef;
        eval {
            local $SIG{ALRM} = sub { die $timeout_msg };
            alarm $timeout;
            $socket->recv($resp, 0x100, 0);
            alarm 0;
        };
        if ($@ eq $timeout_msg) {
            print "timeout.\n";
            next LOGIN;
        }

        if (defined $resp && $resp) {
            my $event = Net::OICQ::ServerEvent->new($resp, $self);
            if (defined $event && $event->seq eq pack('n', $self->{Seq})) {
                $self->{LastSvrAck} = time;
                my $login = $event->login;
                if (defined $login) {
                    if ($login) {
                        print "ok.\n";
                        last LOGIN;
                    } else {
                        $svr = $self->{SvrIP};
                        $port = $self->{SvrPort};
                        $socket = undef;
                        print "redirected.\n";
                        redo LOGIN;
                    }
                } else {
                    die "wrong passwd for $id\n";
                }
            } else {
                print "failed.\n";
            }
        } else {
            print "no response.\n";
        }
        $socket = undef;
        next LOGIN;
    }

    return 0 unless $self->{Key};

    # Make sure we logout when control-C is pressed
    $SIG{INT} = sub { $self->logout; exit 1 };
    # Prepare LogoutPacket for logout
    $self->{LogoutPacket} = $self->build_logout_packet;
    $self->{LastKeepaliveTime} = time;

    return 1;
}

sub send2svr {
    my ($self, $cmd, $data) = @_;
    my $packet = $self->build_packet($cmd, $data);
    my $event = Net::OICQ::ClientEvent->new($packet, $self);
    unless (defined $event) {
        $self->log_t("Failed to create new ClientEvent\n");
        return undef;
    }
    $event->process;
    $self->{Socket}->send($packet);
}

# p2p functions not working yet

sub connect_p2p {
    my ($self, $dstid) = @_;
    defined $self->{Info}->{$dstid} or return;
    my $dst = $self->{Info}->{$dstid}->{Addr};
    defined $dst and $dst =~ /[1-9]/ or return;
    my $socket = IO::Socket::INET->new(
            Proto => 'udp', PeerAddr => $dst
    ) or return;
    $self->{Info}->{$dstid}->{Socket} = $socket;
    return $socket;
}

sub send_p2p {
    my ($self, $dstid, $cmd, $data) = @_;
    defined $self->{Info}->{$dstid} or return;
    my $info = $self->{Info}->{$dstid};
    defined $info->{Socket} or return;
    defined $info->{Key} or return;
    my $packet = $self->build_p2p_packet($dstid, $cmd, $data);
    $info->{Socket}->send($packet);
}

# get_friends_list provided by Chen Peng

sub get_friends_list {
    my ($self, $flag) = @_;
    defined $flag or $flag = "\0\0";
    $self->send2svr('get_friends_list', $flag);
}

sub get_online_friends {
    my ($self) = @_;
    $self->send2svr('get_online_friends', "\x02\0\0\0\0");
}

sub set_mode {
    my ($self, $mode_code) = @_;
    $self->send2svr('set_mode', $mode_code);
}

sub get_user_info {
    my ($self, $id) = @_;
    $self->send2svr('get_user_info', $id);
}

sub update_info {
    my ($self, $hashref) = @_;
    my $info = $self->{MyInfo};
    return unless defined $hashref and defined $info;
    my %new_info;
    # Use all upper-case letters for keys
    foreach my $k (keys %$hashref) {
	$new_info{uc($k)} = $hashref->{$k};
    }
    my @update;
    for (my $i = 1; $i < $#InfoHeader; $i++) {
        my $attr = uc($InfoHeader[$i]);
        push(@update, defined($new_info{$attr}) ? $new_info{$attr} : $info->[$i]);
    }
    $self->send2svr('update_info', join($RS, "", "", @update));
    return 1;
}

sub set_passwd {
    my ($self, $newpw) = @_;
    return unless defined $self->{MyInfo};
    my @info = @{$self->{MyInfo}};
    pop @info; shift @info;
    $self->send2svr('update_info', join($RS, $self->{Passwd}, $newpw, @info));
    return 1;
}

sub accept_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_2', $id.$RS."0");
}

sub reject_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_2', $id.$RS."1");
}

sub add_contact {
    my ($self, $id) = @_;
    $self->send2svr('add_contact_1', "$id");
    $self->{ContactId} = $id;
}

sub add_contact_2 {
    my ($self, $msg) = @_;
    my $mmid = $self->{ContactId};
    return unless defined $mmid;
    $self->send2svr('add_contact_2', "$mmid".$RS."2".$RS."$msg");
    delete $self->{ContactId};
}

sub del_contact {
    my ($self, $id) = @_;
    $self->send2svr('del_contact', "$id");
}

sub forbid_contact {
    my ($self, $id) = @_;
    $self->send2svr('forbid_contact', "$id");
}

# send_msg is also used for auto-reply but does not use AutoMsgPrecursor
# I don't think this is a bug, it is a feature.
sub send_msg {
    my ($self, $dstid, $msg) = @_;
    use bytes;
    $self->log_t("Sent message to $dstid:\n", $msg) if $self->{LogChat} or $self->{Debug};
    my $dstid_ = pack('N', $dstid);
    my $data = $self->{_Id} .
               $dstid_ .
               $SRC_CLIENT .
               $self->{_Id} .
               $dstid_ .
               Digest::MD5::md5($dstid_ . $self->{Key}) .
               "\0\x0b" .
               pack('S', $self->{Seq}) .
               pack('N', time) .
               $MsgPrecursor .
               $msg;
    print "send_msg $dstid:\n", $self->hexdump($data) if $self->{Debug} > 1;
    $self->send2svr('send_msg', $data);
}

sub ack_msg {
    my ($self, $plain) = @_;
    foreach (1..3) {
        $self->send2svr('recv_msg', substr($plain, 0, 12));
    }
}

sub ack_service_msg {
    my ($self, $code, $srcid, $seq) = @_;
    $self->send2svr('ack_service_msg', "$code$FS$srcid$FS$seq");
}

sub keepalive {
    my ($self) = @_;
    $self->send2svr('keep_alive', $self->{Id});
    $self->{LastKeepaliveTime} = time;
}

sub search_user {
    my ($self, $id) = @_;
    $self->send2svr('search_users', join($RS, '0', $id, '-','-','0'));
}

sub list_online_users {
    my ($self, $num) = @_;
    defined $num or $num = 1;
    my $begin = $self->{SearchCount};
    $self->{SearchCount} += $num; 
    my $end = $self->{SearchCount} -1;
    foreach my $p ($begin .. $end) {
        $self->send2svr('search_users', "1".$RS."$p");
    }
}

sub logout {
    my $self = shift;
    defined($self->{LogoutPacket}) && $self->{LogoutPacket} || return;
    my $packet = $self->{LogoutPacket};
    foreach (1..3) {
        $self->{Socket}->send($packet);
    }
}

sub process_packet {
    my ($self, $packet) = @_;

    my $ver = substr($packet, 0, 1);
    if ($ver ne $PROTO_VER) {
        $self->log_t('Unknown protocol version, ', unpack('H*', $ver));
        $self->log($self->hexdump($packet),"\n") if $self->{Debug};
        return;
    }

    my $src = substr($packet, 1, 2);
    my $event;
    if ($src eq $SRC_SERVER) {
        $event = Net::OICQ::ServerEvent->new($packet, $self);
        $self->{LastSvrAck} = time;
    } else {
        $event = Net::OICQ::ClientEvent->new($packet, $self);
    }

    return unless defined $event;

    if ($self->{Debug}) {
        $self->log_t('Seq=', unpack('H*', $event->seq),
                     ', Src=', unpack('H*', $src),
                     ', Cmd=', unpack('H*', $event->cmdcode),
                     " (", $event->cmd, ")");
    }

    return $event->process;
}

1;

__END__

=head1 NAME

Net::OICQ - Perl interface to an OICQ server

=head1 SYNOPSIS

  use Net::OICQ;
  $oicq = new Net::OICQ;

=head1 DESCRIPTION

=head2 EXPORT

None by default.

=head1 AUTHOR

Shufeng Tan <lt>perloicq@yahoo.com<gt>

=head1 SEE ALSO

L<perl>.

=cut
